@@ -12,17 +12,37 @@
 
 
 typedef struct {
-    ngx_queue_t       labels;
+    ngx_queue_t          labels;
 } ngx_js_console_t;
 
 
 typedef struct {
-    njs_str_t         name;
-    uint64_t          time;
-    ngx_queue_t       queue;
+    njs_str_t            name;
+    uint64_t             time;
+    ngx_queue_t          queue;
 } ngx_js_timelabel_t;
 
 
+typedef struct {
+    void                *promise;
+    njs_opaque_value_t   message;
+} ngx_js_rejected_promise_t;
+
+
+#if defined(PATH_MAX)
+#define NGX_MAX_PATH             PATH_MAX
+#else
+#define NGX_MAX_PATH             4096
+#endif
+
+typedef struct {
+    int                 fd;
+    njs_str_t           name;
+    njs_str_t           file;
+    char                path[NGX_MAX_PATH + 1];
+} njs_module_info_t;
+
+
 static njs_int_t ngx_js_ext_build(njs_vm_t *vm, njs_object_prop_t *prop,
     njs_value_t *value, njs_value_t *setval, njs_value_t *retval);
 static njs_int_t ngx_js_ext_conf_file_path(njs_vm_t *vm,
@@ -43,6 +63,13 @@ static njs_int_t ngx_js_ext_console_time(njs_vm_t *vm, njs_value_t *args,
     njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
 static njs_int_t ngx_js_ext_console_time_end(njs_vm_t *vm, njs_value_t *args,
     njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
+static njs_int_t njs_set_timeout(njs_vm_t *vm, njs_value_t *args,
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
+static njs_int_t njs_set_immediate(njs_vm_t *vm, njs_value_t *args,
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
+static njs_int_t njs_clear_timeout(njs_vm_t *vm, njs_value_t *args,
+    njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
+static njs_int_t ngx_js_unhandled_rejection(ngx_js_ctx_t *ctx);
 static void ngx_js_cleanup_vm(void *data);
 
 static njs_int_t ngx_js_core_init(njs_vm_t *vm);
@@ -328,22 +355,63 @@ static njs_int_t      ngx_js_console_proto_id;
 
 
 ngx_int_t
-ngx_js_call(njs_vm_t *vm, ngx_str_t *fname, ngx_log_t *log,
-    njs_opaque_value_t *args, njs_uint_t nargs)
+ngx_js_call(njs_vm_t *vm, njs_function_t *func, njs_value_t *args,
+    njs_uint_t nargs)
 {
-    njs_opaque_value_t  unused;
+    njs_int_t          ret;
+    ngx_str_t          exception;
+    ngx_connection_t  *c;
 
-    return ngx_js_invoke(vm, fname, log, args, nargs, &unused);
+    ret = njs_vm_call(vm, func, args, nargs);
+    if (ret == NJS_ERROR) {
+        ngx_js_exception(vm, &exception);
+
+        c = ngx_external_connection(vm, njs_vm_external_ptr(vm));
+
+        ngx_log_error(NGX_LOG_ERR, c->log, 0,
+                      "js exception: %V", &exception);
+        return NGX_ERROR;
+    }
+
+    for ( ;; ) {
+        ret = njs_vm_execute_pending_job(vm);
+        if (ret <= NJS_OK) {
+            c = ngx_external_connection(vm, njs_vm_external_ptr(vm));
+
+            if (ret == NJS_ERROR) {
+                ngx_js_exception(vm, &exception);
+
+                ngx_log_error(NGX_LOG_ERR, c->log, 0,
+                              "js job exception: %V", &exception);
+                return NGX_ERROR;
+            }
+
+            break;
+        }
+    }
+
+    return NGX_OK;
 }
 
 
 ngx_int_t
-ngx_js_invoke(njs_vm_t *vm, ngx_str_t *fname, ngx_log_t *log,
+ngx_js_name_call(njs_vm_t *vm, ngx_str_t *fname, ngx_log_t *log,
+    njs_opaque_value_t *args, njs_uint_t nargs)
+{
+    njs_opaque_value_t  unused;
+
+    return ngx_js_name_invoke(vm, fname, log, args, nargs, &unused);
+}
+
+
+ngx_int_t
+ngx_js_name_invoke(njs_vm_t *vm, ngx_str_t *fname, ngx_log_t *log,
     njs_opaque_value_t *args, njs_uint_t nargs, njs_opaque_value_t *retval)
 {
     njs_int_t        ret;
     njs_str_t        name;
     ngx_str_t        exception;
+    ngx_js_ctx_t    *ctx;
     njs_function_t  *func;
 
     name.start = fname->data;
@@ -359,7 +427,7 @@ ngx_js_invoke(njs_vm_t *vm, ngx_str_t *fname, ngx_log_t *log,
     ret = njs_vm_invoke(vm, func, njs_value_arg(args), nargs,
                         njs_value_arg(retval));
     if (ret == NJS_ERROR) {
-        ngx_js_retval(vm, NULL, &exception);
+        ngx_js_exception(vm, &exception);
 
         ngx_log_error(NGX_LOG_ERR, log, 0,
                       "js exception: %V", &exception);
@@ -367,33 +435,41 @@ ngx_js_invoke(njs_vm_t *vm, ngx_str_t *fname, ngx_log_t *log,
         return NGX_ERROR;
     }
 
-    ret = njs_vm_run(vm);
-    if (ret == NJS_ERROR) {
-        ngx_js_retval(vm, NULL, &exception);
+    for ( ;; ) {
+        ret = njs_vm_execute_pending_job(vm);
+        if (ret <= NJS_OK) {
+            if (ret == NJS_ERROR) {
+                ngx_js_exception(vm, &exception);
 
-        ngx_log_error(NGX_LOG_ERR, log, 0,
-                      "js exception: %V", &exception);
+                ngx_log_error(NGX_LOG_ERR, log, 0,
+                              "js job exception: %V", &exception);
+                return NGX_ERROR;
+            }
 
+            break;
+        }
+    }
+
+    ctx = ngx_external_ctx(vm, njs_vm_external_ptr(vm));
+
+    if (ngx_js_unhandled_rejection(ctx)) {
+        ngx_js_exception(vm, &exception);
+
+        ngx_log_error(NGX_LOG_ERR, log, 0, "js exception: %V", &exception);
         return NGX_ERROR;
     }
 
-    return (ret == NJS_AGAIN) ? NGX_AGAIN : NGX_OK;
+    return njs_rbtree_is_empty(&ctx->waiting_events) ? NGX_OK : NGX_AGAIN;
 }
 
 
 ngx_int_t
-ngx_js_retval(njs_vm_t *vm, njs_opaque_value_t *retval, ngx_str_t *s)
+ngx_js_exception(njs_vm_t *vm, ngx_str_t *s)
 {
     njs_int_t  ret;
     njs_str_t  str;
 
-    if (retval != NULL && njs_value_is_valid(njs_value_arg(retval))) {
-        ret = njs_vm_value_string(vm, &str, njs_value_arg(retval));
-
-    } else {
-        ret = njs_vm_exception_string(vm, &str);
-    }
-
+    ret = njs_vm_exception_string(vm, &str);
     if (ret != NJS_OK) {
         return NGX_ERROR;
     }
@@ -436,6 +512,78 @@ ngx_js_string(njs_vm_t *vm, njs_value_t *value, njs_str_t *str)
 }
 
 
+static njs_int_t
+njs_function_bind(njs_vm_t *vm, const njs_str_t *name,
+    njs_function_native_t native, njs_bool_t ctor)
+{
+    njs_function_t      *f;
+    njs_opaque_value_t   value;
+
+    f = njs_vm_function_alloc(vm, native, 1, ctor);
+    if (f == NULL) {
+        return NJS_ERROR;
+    }
+
+    njs_value_function_set(njs_value_arg(&value), f);
+
+    return njs_vm_bind(vm, name, njs_value_arg(&value), 1);
+}
+
+
+
+static intptr_t
+ngx_js_event_rbtree_compare(njs_rbtree_node_t *node1, njs_rbtree_node_t *node2)
+{
+    ngx_js_event_t  *ev1, *ev2;
+
+    ev1 = (ngx_js_event_t *) ((u_char *) node1
+                              - offsetof(ngx_js_event_t, node));
+    ev2 = (ngx_js_event_t *) ((u_char *) node2
+                              - offsetof(ngx_js_event_t, node));
+
+    if (ev1->fd < ev2->fd) {
+        return -1;
+    }
+
+    if (ev1->fd > ev2->fd) {
+        return 1;
+    }
+
+    return 0;
+}
+
+
+void
+ngx_js_ctx_init(ngx_js_ctx_t *ctx)
+{
+    ctx->event_id = 0;
+    njs_rbtree_init(&ctx->waiting_events, ngx_js_event_rbtree_compare);
+}
+
+
+void
+ngx_js_ctx_destroy(ngx_js_ctx_t *ctx)
+{
+    ngx_js_event_t     *event;
+    njs_rbtree_node_t  *node;
+
+    node = njs_rbtree_min(&ctx->waiting_events);
+
+    while (njs_rbtree_is_there_successor(&ctx->waiting_events, node)) {
+        event = (ngx_js_event_t *) ((u_char *) node
+                                    - offsetof(ngx_js_event_t, node));
+
+        if (event->destructor != NULL) {
+            event->destructor(njs_vm_external_ptr(event->vm), event);
+        }
+
+        node = njs_rbtree_node_successor(&ctx->waiting_events, node);
+    }
+
+    njs_vm_destroy(ctx->vm);
+}
+
+
 static njs_int_t
 ngx_js_core_init(njs_vm_t *vm)
 {
@@ -443,6 +591,10 @@ ngx_js_core_init(njs_vm_t *vm)
     njs_str_t           name;
     njs_opaque_value_t  value;
 
+    static const njs_str_t  set_timeout = njs_str("setTimeout");
+    static const njs_str_t  set_immediate = njs_str("setImmediate");
+    static const njs_str_t  clear_timeout = njs_str("clearTimeout");
+
     proto_id = njs_vm_external_prototype(vm, ngx_js_ext_core,
                                          njs_nitems(ngx_js_ext_core));
     if (proto_id < 0) {
@@ -482,6 +634,21 @@ ngx_js_core_init(njs_vm_t *vm)
         return NJS_ERROR;
     }
 
+    ret = njs_function_bind(vm, &set_timeout, njs_set_timeout, 1);
+    if (ret != NJS_OK) {
+        return NJS_ERROR;
+    }
+
+    ret = njs_function_bind(vm, &set_immediate, njs_set_immediate, 1);
+    if (ret != NJS_OK) {
+        return NJS_ERROR;
+    }
+
+    ret = njs_function_bind(vm, &clear_timeout, njs_clear_timeout, 1);
+    if (ret != NJS_OK) {
+        return NJS_ERROR;
+    }
+
     return NJS_OK;
 }
 
@@ -501,7 +668,7 @@ ngx_js_ext_string(njs_vm_t *vm, njs_object_prop_t *prop, njs_value_t *value,
 
     field = (ngx_str_t *) (p + njs_vm_prop_magic32(prop));
 
-    return njs_vm_value_string_set(vm, retval, field->data, field->len);
+    return njs_vm_value_string_create(vm, retval, field->data, field->len);
 }
 
 
@@ -578,15 +745,15 @@ njs_int_t
 ngx_js_ext_build(njs_vm_t *vm, njs_object_prop_t *prop, njs_value_t *value,
     njs_value_t *setval, njs_value_t *retval)
 {
-    return njs_vm_value_string_set(vm, retval,
+    return njs_vm_value_string_create(vm, retval,
 #ifdef NGX_BUILD
-                                   (u_char *) NGX_BUILD,
-                                   njs_strlen(NGX_BUILD)
+                                      (u_char *) NGX_BUILD,
+                                      njs_strlen(NGX_BUILD)
 #else
-                                   (u_char *) "",
-                                   0
+                                      (u_char *) "",
+                                      0
 #endif
-                                   );
+                                     );
 }
 
 
@@ -594,8 +761,8 @@ njs_int_t
 ngx_js_ext_conf_file_path(njs_vm_t *vm, njs_object_prop_t *prop,
     njs_value_t *value, njs_value_t *setval, njs_value_t *retval)
 {
-    return njs_vm_value_string_set(vm, retval, ngx_cycle->conf_file.data,
-                                   ngx_cycle->conf_file.len);
+    return njs_vm_value_string_create(vm, retval, ngx_cycle->conf_file.data,
+                                      ngx_cycle->conf_file.len);
 }
 
 
@@ -603,8 +770,8 @@ njs_int_t
 ngx_js_ext_conf_prefix(njs_vm_t *vm, njs_object_prop_t *prop,
     njs_value_t *value, njs_value_t *setval, njs_value_t *retval)
 {
-    return njs_vm_value_string_set(vm, retval, ngx_cycle->conf_prefix.data,
-                                   ngx_cycle->conf_prefix.len);
+    return njs_vm_value_string_create(vm, retval, ngx_cycle->conf_prefix.data,
+                                      ngx_cycle->conf_prefix.len);
 }
 
 
@@ -612,8 +779,8 @@ njs_int_t
 ngx_js_ext_error_log_path(njs_vm_t *vm, njs_object_prop_t *prop,
     njs_value_t *value, njs_value_t *setval, njs_value_t *retval)
 {
-    return njs_vm_value_string_set(vm, retval, ngx_cycle->error_log.data,
-                                   ngx_cycle->error_log.len);
+    return njs_vm_value_string_create(vm, retval, ngx_cycle->error_log.data,
+                                      ngx_cycle->error_log.len);
 }
 
 
@@ -621,8 +788,8 @@ njs_int_t
 ngx_js_ext_prefix(njs_vm_t *vm, njs_object_prop_t *prop, njs_value_t *value,
     njs_value_t *setval, njs_value_t *retval)
 {
-    return njs_vm_value_string_set(vm, retval, ngx_cycle->prefix.data,
-                                   ngx_cycle->prefix.len);
+    return njs_vm_value_string_create(vm, retval, ngx_cycle->prefix.data,
+                                      ngx_cycle->prefix.len);
 }
 
 
@@ -630,8 +797,8 @@ njs_int_t
 ngx_js_ext_version(njs_vm_t *vm, njs_object_prop_t *prop, njs_value_t *value,
     njs_value_t *setval, njs_value_t *retval)
 {
-    return njs_vm_value_string_set(vm, retval, (u_char *) NGINX_VERSION,
-                                   njs_strlen(NGINX_VERSION));
+    return njs_vm_value_string_create(vm, retval, (u_char *) NGINX_VERSION,
+                                      njs_strlen(NGINX_VERSION));
 }
 
 
@@ -648,11 +815,10 @@ njs_int_t
 ngx_js_ext_log(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     njs_index_t magic, njs_value_t *retval)
 {
-    char             *p;
-    ngx_int_t        lvl;
-    njs_str_t        msg;
-    njs_uint_t       n;
-    njs_log_level_t  level;
+    char        *p;
+    ngx_int_t   lvl;
+    njs_str_t   msg;
+    njs_uint_t  n, level;
 
     p = njs_vm_external(vm, NJS_PROTO_ID_ANY, njs_argument(args, 0));
     if (p == NULL) {
@@ -660,7 +826,7 @@ ngx_js_ext_log(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    level = (njs_log_level_t) magic & NGX_JS_LOG_MASK;
+    level = magic & NGX_JS_LOG_MASK;
 
     if (level == 0) {
         if (ngx_js_integer(vm, njs_arg(args, nargs, 1), &lvl) != NGX_OK) {
@@ -753,7 +919,8 @@ ngx_js_ext_console_time(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         label = ngx_queue_data(q, ngx_js_timelabel_t, queue);
 
         if (njs_strstr_eq(&name, &label->name)) {
-            njs_vm_log(vm, "Timer \"%V\" already exists.\n", &name);
+            ngx_js_log(vm, njs_vm_external_ptr(vm), NGX_LOG_INFO,
+                       "Timer \"%V\" already exists.", &name);
             njs_value_undefined_set(retval);
             return NJS_OK;
         }
@@ -849,7 +1016,8 @@ ngx_js_ext_console_time_end(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     ms = ns / 1000000;
     ns = ns % 1000000;
 
-    njs_vm_log(vm, "%V: %uL.%06uLms\n", &name, ms, ns);
+    ngx_js_log(vm, njs_vm_external_ptr(vm), NGX_LOG_INFO, "%V: %uL.%06uLms",
+               &name, ms, ns);
 
     njs_value_undefined_set(retval);
 
@@ -857,7 +1025,155 @@ ngx_js_ext_console_time_end(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
 not_found:
 
-    njs_vm_log(vm, "Timer \"%V\" doesn't exist.\n", &name);
+    ngx_js_log(vm, njs_vm_external_ptr(vm), NGX_LOG_INFO,
+               "Timer \"%V\" doesn't exist.", &name);
+
+    njs_value_undefined_set(retval);
+
+    return NJS_OK;
+}
+
+
+static void
+ngx_js_timer_handler(ngx_event_t *ev)
+{
+    njs_vm_t        *vm;
+    ngx_int_t        rc;
+    ngx_js_ctx_t    *ctx;
+    ngx_js_event_t  *event;
+
+    event = (ngx_js_event_t *) ((u_char *) ev - offsetof(ngx_js_event_t, ev));
+
+    vm = event->vm;
+
+    rc = ngx_js_call(vm, event->function, event->args, event->nargs);
+
+    ctx = ngx_external_ctx(vm, njs_vm_external_ptr(vm));
+    ngx_js_del_event(ctx, event);
+
+    ngx_external_event_finalize(vm)(njs_vm_external_ptr(vm), rc);
+}
+
+
+static void
+ngx_js_clear_timer(njs_external_ptr_t external, ngx_js_event_t *event)
+{
+    if (event->ev.timer_set) {
+        ngx_del_timer(&event->ev);
+    }
+}
+
+
+static njs_int_t
+njs_set_timer(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
+    njs_index_t unused, njs_bool_t immediate, njs_value_t *retval)
+{
+    uint64_t           delay;
+    njs_uint_t         n;
+    ngx_js_ctx_t      *ctx;
+    ngx_js_event_t    *event;
+    ngx_connection_t  *c;
+
+    if (njs_slow_path(nargs < 2)) {
+        njs_vm_type_error(vm, "too few arguments");
+        return NJS_ERROR;
+    }
+
+    if (njs_slow_path(!njs_value_is_function(njs_argument(args, 1)))) {
+        njs_vm_type_error(vm, "first arg must be a function");
+        return NJS_ERROR;
+    }
+
+    delay = 0;
+
+    if (!immediate && nargs >= 3
+        && njs_value_is_number(njs_argument(args, 2)))
+    {
+        delay = njs_value_number(njs_argument(args, 2));
+    }
+
+    n = immediate ? 2 : 3;
+    nargs = (nargs >= n) ? nargs - n : 0;
+
+    event = njs_mp_zalloc(njs_vm_memory_pool(vm),
+                          sizeof(ngx_js_event_t)
+                          + sizeof(njs_opaque_value_t) * nargs);
+    if (njs_slow_path(event == NULL)) {
+        njs_vm_memory_error(vm);
+        return NJS_ERROR;
+    }
+
+    event->vm = vm;
+    event->function = njs_value_function(njs_argument(args, 1));
+    event->nargs = nargs;
+    event->args = (njs_value_t *) ((u_char *) event + sizeof(ngx_js_event_t));
+    event->destructor = ngx_js_clear_timer;
+
+    ctx = ngx_external_ctx(vm, njs_vm_external_ptr(vm));
+    event->fd = ctx->event_id++;
+
+    c = ngx_external_connection(vm, njs_vm_external_ptr(vm));
+
+    event->ev.log = c->log;
+    event->ev.data = event;
+    event->ev.handler = ngx_js_timer_handler;
+
+    if (event->nargs != 0) {
+        memcpy(event->args, njs_argument(args, n),
+               sizeof(njs_opaque_value_t) * event->nargs);
+    }
+
+    ngx_js_add_event(ctx, event);
+
+    ngx_add_timer(&event->ev, delay);
+
+    njs_value_number_set(retval, event->fd);
+
+    return NJS_OK;
+}
+
+
+static njs_int_t
+njs_set_timeout(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
+    njs_index_t unused, njs_value_t *retval)
+{
+    return njs_set_timer(vm, args, nargs, unused, 0, retval);
+}
+
+
+static njs_int_t
+njs_set_immediate(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
+    njs_index_t unused, njs_value_t *retval)
+{
+    return njs_set_timer(vm, args, nargs, unused, 1, retval);
+}
+
+
+static njs_int_t
+njs_clear_timeout(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
+    njs_index_t unused, njs_value_t *retval)
+{
+    ngx_js_ctx_t       *ctx;
+    ngx_js_event_t      event_lookup, *event;
+    njs_rbtree_node_t  *rb;
+
+    if (nargs < 2 || !njs_value_is_number(njs_argument(args, 1))) {
+        njs_value_undefined_set(retval);
+        return NJS_OK;
+    }
+
+    ctx = ngx_external_ctx(vm, njs_vm_external_ptr(vm));
+    event_lookup.fd = njs_value_number(njs_argument(args, 1));
+
+    rb = njs_rbtree_find(&ctx->waiting_events, &event_lookup.node);
+    if (njs_slow_path(rb == NULL)) {
+        njs_vm_internal_error(vm, "failed to find timer");
+        return NJS_ERROR;
+    }
+
+    event = (ngx_js_event_t *) ((u_char *) rb - offsetof(ngx_js_event_t, node));
+
+    ngx_js_del_event(ctx, event);
 
     njs_value_undefined_set(retval);
 
@@ -866,7 +1182,23 @@ not_found:
 
 
 void
-ngx_js_logger(njs_vm_t *vm, njs_external_ptr_t external, njs_log_level_t level,
+ngx_js_log(njs_vm_t *vm, njs_external_ptr_t external, ngx_uint_t level,
+    const char *fmt, ...)
+{
+    u_char   *p;
+    va_list   args;
+    u_char    buf[NGX_MAX_ERROR_STR];
+
+    va_start(args, fmt);
+    p = njs_vsprintf(buf, buf + sizeof(buf), fmt, args);
+    va_end(args);
+
+    ngx_js_logger(vm, external, level, buf, p - buf);
+}
+
+
+void
+ngx_js_logger(njs_vm_t *vm, njs_external_ptr_t external, ngx_uint_t level,
     const u_char *start, size_t length)
 {
     ngx_log_t           *log;
@@ -888,7 +1220,7 @@ ngx_js_logger(njs_vm_t *vm, njs_external_ptr_t external, njs_log_level_t level,
         log = ngx_cycle->log;
     }
 
-    ngx_log_error((ngx_uint_t) level, log, 0, "js: %*s", length, start);
+    ngx_log_error(level, log, 0, "js: %*s", length, start);
 
     if (external != NULL) {
         log->handler = handler;
@@ -1350,6 +1682,309 @@ ngx_js_merge_vm(ngx_conf_t *cf, ngx_js_loc_conf_t *conf,
 }
 
 
+static void
+ngx_js_rejection_tracker(njs_vm_t *vm, njs_external_ptr_t unused,
+    njs_bool_t is_handled, njs_value_t *promise, njs_value_t *reason)
+{
+    void                       *promise_obj;
+    uint32_t                    i, length;
+    ngx_js_ctx_t               *ctx;
+    ngx_js_rejected_promise_t  *rejected_promise;
+
+    ctx = ngx_external_ctx(vm, njs_vm_external_ptr(vm));
+
+    if (is_handled && ctx->rejected_promises != NULL) {
+        rejected_promise = ctx->rejected_promises->start;
+        length = ctx->rejected_promises->items;
+
+        promise_obj = njs_value_ptr(promise);
+
+        for (i = 0; i < length; i++) {
+            if (rejected_promise[i].promise == promise_obj) {
+                njs_arr_remove(ctx->rejected_promises,
+                               &rejected_promise[i]);
+
+                break;
+            }
+        }
+
+        return;
+    }
+
+    if (ctx->rejected_promises == NULL) {
+        ctx->rejected_promises = njs_arr_create(njs_vm_memory_pool(vm), 4,
+                                             sizeof(ngx_js_rejected_promise_t));
+        if (njs_slow_path(ctx->rejected_promises == NULL)) {
+            return;
+        }
+    }
+
+    rejected_promise = njs_arr_add(ctx->rejected_promises);
+    if (njs_slow_path(rejected_promise == NULL)) {
+        return;
+    }
+
+    rejected_promise->promise = njs_value_ptr(promise);
+    njs_value_assign(&rejected_promise->message, reason);
+}
+
+
+static njs_int_t
+ngx_js_module_path(const ngx_str_t *dir, njs_module_info_t *info)
+{
+    char        *p;
+    size_t      length;
+    njs_bool_t  trail;
+    char        src[NGX_MAX_PATH + 1];
+
+    trail = 0;
+    length = info->name.length;
+
+    if (dir != NULL) {
+        length += dir->len;
+
+        if (length == 0 || dir->len == 0) {
+            return NJS_DECLINED;
+        }
+
+        trail = (dir->data[dir->len - 1] != '/');
+
+        if (trail) {
+            length++;
+        }
+    }
+
+    if (njs_slow_path(length > NGX_MAX_PATH)) {
+        return NJS_ERROR;
+    }
+
+    p = &src[0];
+
+    if (dir != NULL) {
+        p = (char *) njs_cpymem(p, dir->data, dir->len);
+
+        if (trail) {
+            *p++ = '/';
+        }
+    }
+
+    p = (char *) njs_cpymem(p, info->name.start, info->name.length);
+    *p = '\0';
+
+    p = realpath(&src[0], &info->path[0]);
+    if (p == NULL) {
+        return NJS_DECLINED;
+    }
+
+    info->fd = open(&info->path[0], O_RDONLY);
+    if (info->fd < 0) {
+        return NJS_DECLINED;
+    }
+
+    info->file.start = (u_char *) &info->path[0];
+    info->file.length = njs_strlen(info->file.start);
+
+    return NJS_OK;
+}
+
+
+static njs_int_t
+ngx_js_module_lookup(ngx_js_loc_conf_t *conf, njs_module_info_t *info)
+{
+    njs_int_t   ret;
+    ngx_str_t   *path;
+    njs_uint_t  i;
+
+    if (info->name.start[0] == '/') {
+        return ngx_js_module_path(NULL, info);
+    }
+
+    ret = ngx_js_module_path(&conf->cwd, info);
+
+    if (ret != NJS_DECLINED) {
+        return ret;
+    }
+
+    ret = ngx_js_module_path((const ngx_str_t *) &ngx_cycle->conf_prefix, info);
+
+    if (ret != NJS_DECLINED) {
+        return ret;
+    }
+
+    if (conf->paths == NGX_CONF_UNSET_PTR) {
+        return NJS_DECLINED;
+    }
+
+    path = conf->paths->elts;
+
+    for (i = 0; i < conf->paths->nelts; i++) {
+        ret = ngx_js_module_path(&path[i], info);
+
+        if (ret != NJS_DECLINED) {
+            return ret;
+        }
+    }
+
+    return NJS_DECLINED;
+}
+
+
+static njs_int_t
+ngx_js_module_read(njs_mp_t *mp, int fd, njs_str_t *text)
+{
+    ssize_t      n;
+    struct stat  sb;
+
+    text->start = NULL;
+
+    if (fstat(fd, &sb) == -1) {
+        goto fail;
+    }
+
+    if (!S_ISREG(sb.st_mode)) {
+        goto fail;
+    }
+
+    text->length = sb.st_size;
+
+    text->start = njs_mp_alloc(mp, text->length);
+    if (text->start == NULL) {
+        goto fail;
+    }
+
+    n = read(fd, text->start, sb.st_size);
+
+    if (n < 0 || n != sb.st_size) {
+        goto fail;
+    }
+
+    return NJS_OK;
+
+fail:
+
+    if (text->start != NULL) {
+        njs_mp_free(mp, text->start);
+    }
+
+    return NJS_ERROR;
+}
+
+
+static void
+ngx_js_file_dirname(const njs_str_t *path, ngx_str_t *name)
+{
+    const u_char  *p, *end;
+
+    if (path->length == 0) {
+        goto current_dir;
+    }
+
+    p = path->start + path->length - 1;
+
+    /* Stripping basename. */
+
+    while (p >= path->start && *p != '/') { p--; }
+
+    end = p + 1;
+
+    if (end == path->start) {
+        goto current_dir;
+    }
+
+    /* Stripping trailing slashes. */
+
+    while (p >= path->start && *p == '/') { p--; }
+
+    p++;
+
+    if (p == path->start) {
+        p = end;
+    }
+
+    name->data = path->start;
+    name->len = p - path->start;
+
+    return;
+
+current_dir:
+
+    ngx_str_set(name, ".");
+}
+
+
+static njs_int_t
+ngx_js_set_cwd(njs_vm_t *vm, ngx_js_loc_conf_t *conf, njs_str_t *path)
+{
+    ngx_str_t  cwd;
+
+    ngx_js_file_dirname(path, &cwd);
+
+    conf->cwd.data = njs_mp_alloc(njs_vm_memory_pool(vm), cwd.len);
+    if (conf->cwd.data == NULL) {
+        return NJS_ERROR;
+    }
+
+    memcpy(conf->cwd.data, cwd.data, cwd.len);
+    conf->cwd.len = cwd.len;
+
+    return NJS_OK;
+}
+
+
+static njs_mod_t *
+ngx_js_module_loader(njs_vm_t *vm, njs_external_ptr_t external, njs_str_t *name)
+{
+    u_char             *start;
+    njs_int_t           ret;
+    njs_str_t           text;
+    ngx_str_t           prev_cwd;
+    njs_mod_t          *module;
+    ngx_js_loc_conf_t  *conf;
+    njs_module_info_t   info;
+
+    conf = external;
+
+    njs_memzero(&info, sizeof(njs_module_info_t));
+
+    info.name = *name;
+
+    ret = ngx_js_module_lookup(conf, &info);
+    if (njs_slow_path(ret != NJS_OK)) {
+        return NULL;
+    }
+
+    ret = ngx_js_module_read(njs_vm_memory_pool(vm), info.fd, &text);
+
+    (void) close(info.fd);
+
+    if (ret != NJS_OK) {
+        njs_vm_internal_error(vm, "while reading \"%V\" module", &info.file);
+        return NULL;
+    }
+
+    prev_cwd = conf->cwd;
+
+    ret = ngx_js_set_cwd(vm, conf, &info.file);
+    if (ret != NJS_OK) {
+        njs_vm_internal_error(vm, "while setting cwd for \"%V\" module",
+                              &info.file);
+        return NULL;
+    }
+
+    start = text.start;
+
+    module = njs_vm_compile_module(vm, &info.file, &start,
+                                   &text.start[text.length]);
+
+    njs_mp_free(njs_vm_memory_pool(vm), conf->cwd.data);
+    conf->cwd = prev_cwd;
+
+    njs_mp_free(njs_vm_memory_pool(vm), text.start);
+
+    return module;
+}
+
+
 ngx_int_t
 ngx_js_init_conf_vm(ngx_conf_t *cf, ngx_js_loc_conf_t *conf,
     njs_vm_opt_t *options)
@@ -1358,7 +1993,7 @@ ngx_js_init_conf_vm(ngx_conf_t *cf, ngx_js_loc_conf_t *conf,
     u_char               *start, *end, *p;
     ngx_str_t            *m, file;
     njs_int_t             rc;
-    njs_str_t             text, path;
+    njs_str_t             text;
     ngx_uint_t            i;
     njs_value_t          *value;
     ngx_pool_cleanup_t   *cln;
@@ -1427,15 +2062,17 @@ ngx_js_init_conf_vm(ngx_conf_t *cf, ngx_js_loc_conf_t *conf,
     cln->handler = ngx_js_cleanup_vm;
     cln->data = conf;
 
-    path.start = ngx_cycle->conf_prefix.data;
-    path.length = ngx_cycle->conf_prefix.len;
+    njs_vm_set_rejection_tracker(conf->vm, ngx_js_rejection_tracker,
+                                 NULL);
 
-    rc = njs_vm_add_path(conf->vm, &path);
+    rc = ngx_js_set_cwd(conf->vm, conf, &options->file);
     if (rc != NJS_OK) {
-        ngx_log_error(NGX_LOG_EMERG, cf->log, 0, "failed to add \"js_path\"");
+        ngx_log_error(NGX_LOG_EMERG, cf->log, 0, "failed to set cwd");
         return NGX_ERROR;
     }
 
+    njs_vm_set_module_loader(conf->vm, ngx_js_module_loader, conf);
+
     if (conf->paths != NGX_CONF_UNSET_PTR) {
         m = conf->paths->elts;
 
@@ -1443,16 +2080,6 @@ ngx_js_init_conf_vm(ngx_conf_t *cf, ngx_js_loc_conf_t *conf,
             if (ngx_conf_full_name(cf->cycle, &m[i], 1) != NGX_OK) {
                 return NGX_ERROR;
             }
-
-            path.start = m[i].data;
-            path.length = m[i].len;
-
-            rc = njs_vm_add_path(conf->vm, &path);
-            if (rc != NJS_OK) {
-                ngx_log_error(NGX_LOG_EMERG, cf->log, 0,
-                              "failed to add \"js_path\"");
-                return NGX_ERROR;
-            }
         }
     }
 
@@ -1499,6 +2126,36 @@ ngx_js_init_conf_vm(ngx_conf_t *cf, ngx_js_loc_conf_t *conf,
 }
 
 
+static njs_int_t
+ngx_js_unhandled_rejection(ngx_js_ctx_t *ctx)
+{
+    njs_int_t                   ret;
+    njs_str_t                   message;
+    ngx_js_rejected_promise_t  *rejected_promise;
+
+    if (ctx->rejected_promises == NULL
+        || ctx->rejected_promises->items == 0)
+    {
+        return 0;
+    }
+
+    rejected_promise = ctx->rejected_promises->start;
+
+    ret = njs_vm_value_to_string(ctx->vm, &message,
+                                 njs_value_arg(&rejected_promise->message));
+    if (njs_slow_path(ret != NJS_OK)) {
+        return -1;
+    }
+
+    njs_vm_error(ctx->vm, "unhandled promise rejection: %V", &message);
+
+    njs_arr_destroy(ctx->rejected_promises);
+    ctx->rejected_promises = NULL;
+
+    return 1;
+}
+
+
 static void
 ngx_js_cleanup_vm(void *data)
 {
