@@ -12,7 +12,10 @@
 
 #include <ngx_config.h>
 #include <ngx_core.h>
+#include <ngx_event.h>
 #include <njs.h>
+#include <njs_rbtree.h>
+#include <njs_arr.h>
 #include "ngx_js_fetch.h"
 #include "ngx_js_shared_dict.h"
 
@@ -31,9 +34,13 @@
 #define ngx_js_buffer_type(btype) ((btype) & ~NGX_JS_DEPRECATED)
 
 
+typedef struct ngx_js_event_s ngx_js_event_t;
+typedef struct ngx_js_dict_s  ngx_js_dict_t;
+typedef struct ngx_js_ctx_s  ngx_js_ctx_t;
+
+
 typedef ngx_pool_t *(*ngx_external_pool_pt)(njs_vm_t *vm, njs_external_ptr_t e);
-typedef void (*ngx_js_event_handler_pt)(njs_external_ptr_t e,
-    njs_vm_event_t vm_event, njs_value_t *args, njs_uint_t nargs);
+typedef void (*ngx_js_event_finalize_pt)(njs_external_ptr_t e, ngx_int_t rc);
 typedef ngx_resolver_t *(*ngx_external_resolver_pt)(njs_vm_t *vm,
     njs_external_ptr_t e);
 typedef ngx_msec_t (*ngx_external_timeout_pt)(njs_vm_t *vm,
@@ -43,10 +50,10 @@ typedef ngx_flag_t (*ngx_external_flag_pt)(njs_vm_t *vm,
 typedef ngx_flag_t (*ngx_external_size_pt)(njs_vm_t *vm,
     njs_external_ptr_t e);
 typedef ngx_ssl_t *(*ngx_external_ssl_pt)(njs_vm_t *vm, njs_external_ptr_t e);
+typedef ngx_js_ctx_t *(*ngx_js_external_ctx_pt)(njs_vm_t *vm,
+    njs_external_ptr_t e);
 
 
-typedef struct ngx_js_dict_s  ngx_js_dict_t;
-
 typedef struct {
     ngx_str_t              name;
     ngx_str_t              path;
@@ -55,6 +62,20 @@ typedef struct {
 } ngx_js_named_path_t;
 
 
+struct ngx_js_event_s {
+    njs_vm_t            *vm;
+    njs_function_t      *function;
+    njs_value_t         *args;
+    ngx_socket_t         fd;
+    NJS_RBTREE_NODE     (node);
+    njs_uint_t           nargs;
+    void               (*destructor)(njs_external_ptr_t external,
+                                     ngx_js_event_t *event);
+    ngx_event_t          ev;
+    void                *data;
+};
+
+
 #define NGX_JS_COMMON_MAIN_CONF                                               \
     ngx_js_dict_t         *dicts;                                             \
     ngx_array_t           *periodics                                          \
@@ -62,6 +83,7 @@ typedef struct {
 
 #define _NGX_JS_COMMON_LOC_CONF                                               \
     njs_vm_t              *vm;                                                \
+    ngx_str_t              cwd;                                               \
     ngx_array_t           *imports;                                           \
     ngx_array_t           *paths;                                             \
                                                                               \
@@ -89,6 +111,27 @@ typedef struct {
 #endif
 
 
+#define NGX_JS_COMMON_CTX                                                     \
+    njs_vm_t              *vm;                                                \
+    njs_arr_t             *rejected_promises;                                 \
+    njs_rbtree_t           waiting_events;                                    \
+    ngx_socket_t           event_id
+
+
+#define ngx_js_add_event(ctx, event)                                          \
+    njs_rbtree_insert(&(ctx)->waiting_events, &(event)->node)
+
+
+#define ngx_js_del_event(ctx, event)                                          \
+    do {                                                                      \
+        if ((event)->destructor) {                                            \
+            (event)->destructor(njs_vm_external_ptr((event)->vm), event);     \
+        }                                                                     \
+                                                                              \
+        njs_rbtree_delete(&(ctx)->waiting_events, &(event)->node);            \
+    } while (0)
+
+
 typedef struct {
     NGX_JS_COMMON_MAIN_CONF;
 } ngx_js_main_conf_t;
@@ -99,6 +142,11 @@ typedef struct {
 } ngx_js_loc_conf_t;
 
 
+struct ngx_js_ctx_s {
+    NGX_JS_COMMON_CTX;
+};
+
+
 #define ngx_external_connection(vm, e)                                        \
     (*((ngx_connection_t **) ((u_char *) (e) + njs_vm_meta(vm, 0))))
 #define ngx_external_pool(vm, e)                                              \
@@ -107,8 +155,8 @@ typedef struct {
     ((ngx_external_resolver_pt) njs_vm_meta(vm, 2))(vm, e)
 #define ngx_external_resolver_timeout(vm, e)                                  \
     ((ngx_external_timeout_pt) njs_vm_meta(vm, 3))(vm, e)
-#define ngx_external_event_handler(vm, e)                                     \
-    ((ngx_js_event_handler_pt) njs_vm_meta(vm, 4))
+#define ngx_external_event_finalize(vm) \
+    ((ngx_js_event_finalize_pt) njs_vm_meta(vm, 4))
 #define ngx_external_ssl(vm, e)                                               \
     ((ngx_external_ssl_pt) njs_vm_meta(vm, 5))(vm, e)
 #define ngx_external_ssl_verify(vm, e)                                        \
@@ -122,24 +170,35 @@ typedef struct {
 #define NGX_JS_MAIN_CONF_INDEX  10
 #define ngx_main_conf(vm)                                                     \
 	((ngx_js_main_conf_t *) njs_vm_meta(vm, NGX_JS_MAIN_CONF_INDEX))
+#define ngx_external_ctx(vm, e) \
+    ((ngx_js_external_ctx_pt) njs_vm_meta(vm, 11))(vm, e)
 
 
 #define ngx_js_prop(vm, type, value, start, len)                              \
-    ((type == NGX_JS_STRING) ? njs_vm_value_string_set(vm, value, start, len) \
+    ((type == NGX_JS_STRING) ? njs_vm_value_string_create(vm, value, start, len) \
                              : njs_vm_value_buffer_set(vm, value, start, len))
 
 
-ngx_int_t ngx_js_call(njs_vm_t *vm, ngx_str_t *fname, ngx_log_t *log,
+#define ngx_vm_pending(ctx)                                                   \
+    (njs_vm_pending((ctx)->vm) || !njs_rbtree_is_empty(&(ctx)->waiting_events))
+
+
+void ngx_js_ctx_init(ngx_js_ctx_t *ctx);
+void ngx_js_ctx_destroy(ngx_js_ctx_t *ctx);
+ngx_int_t ngx_js_call(njs_vm_t *vm, njs_function_t *func, njs_value_t *args,
+    njs_uint_t nargs);
+ngx_int_t ngx_js_name_call(njs_vm_t *vm, ngx_str_t *fname, ngx_log_t *log,
     njs_opaque_value_t *args, njs_uint_t nargs);
-ngx_int_t ngx_js_invoke(njs_vm_t *vm, ngx_str_t *fname, ngx_log_t *log,
+ngx_int_t ngx_js_name_invoke(njs_vm_t *vm, ngx_str_t *fname, ngx_log_t *log,
     njs_opaque_value_t *args, njs_uint_t nargs, njs_opaque_value_t *retval);
-ngx_int_t ngx_js_retval(njs_vm_t *vm, njs_opaque_value_t *retval,
-    ngx_str_t *s);
+ngx_int_t ngx_js_exception(njs_vm_t *vm, ngx_str_t *s);
 
 njs_int_t ngx_js_ext_log(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     njs_index_t level, njs_value_t *retval);
-void ngx_js_logger(njs_vm_t *vm, njs_external_ptr_t external,
-    njs_log_level_t level, const u_char *start, size_t length);
+void ngx_js_log(njs_vm_t *vm, njs_external_ptr_t external,
+    ngx_uint_t level, const char *fmt, ...);
+void ngx_js_logger(njs_vm_t *vm, njs_external_ptr_t external, ngx_uint_t level,
+    const u_char *start, size_t length);
 char * ngx_js_import(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
 char * ngx_js_preload_object(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);
 ngx_int_t ngx_js_init_preload_vm(ngx_conf_t *cf, ngx_js_loc_conf_t *conf);
