@@ -85,15 +85,16 @@ static ngx_js_dict_node_t *ngx_js_dict_lookup(ngx_js_dict_t *dict,
 #define NGX_JS_DICT_FLAG_MUST_NOT_EXIST   2
 
 static ngx_int_t ngx_js_dict_set(njs_vm_t *vm, ngx_js_dict_t *dict,
-    njs_str_t *key, njs_value_t *value, unsigned flags);
+    njs_str_t *key, njs_value_t *value, ngx_msec_t timeout, unsigned flags);
 static ngx_int_t ngx_js_dict_add(ngx_js_dict_t *dict, njs_str_t *key,
-    njs_value_t *value, ngx_msec_t now);
+    njs_value_t *value, ngx_msec_t timeout, ngx_msec_t now);
 static ngx_int_t ngx_js_dict_update(ngx_js_dict_t *dict,
-    ngx_js_dict_node_t *node, njs_value_t *value, ngx_msec_t now);
+    ngx_js_dict_node_t *node, njs_value_t *value, ngx_msec_t timeout,
+    ngx_msec_t now);
 static ngx_int_t ngx_js_dict_get(njs_vm_t *vm, ngx_js_dict_t *dict,
     njs_str_t *key, njs_value_t *retval);
 static ngx_int_t ngx_js_dict_incr(ngx_js_dict_t *dict, njs_str_t *key,
-    njs_value_t *delta, njs_value_t *init, double *value);
+    njs_value_t *delta, njs_value_t *init, double *value, ngx_msec_t timeout);
 static ngx_int_t ngx_js_dict_delete(njs_vm_t *vm, ngx_js_dict_t *dict,
     njs_str_t *key, njs_value_t *retval);
 static ngx_int_t ngx_js_dict_copy_value_locked(njs_vm_t *vm,
@@ -423,8 +424,8 @@ njs_js_ext_global_shared_keys(njs_vm_t *vm, njs_value_t *unused,
             return NJS_ERROR;
         }
 
-        rc = njs_vm_value_string_set(vm, value, shm_zone->shm.name.data,
-                                     shm_zone->shm.name.len);
+        rc = njs_vm_value_string_create(vm, value, shm_zone->shm.name.data,
+                                        shm_zone->shm.name.len);
         if (rc != NJS_OK) {
             return NJS_ERROR;
         }
@@ -698,8 +699,8 @@ njs_js_ext_shared_dict_keys(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
             goto fail;
         }
 
-        rc = njs_vm_value_string_set(vm, value, node->sn.str.data,
-                                     node->sn.str.len);
+        rc = njs_vm_value_string_create(vm, value, node->sn.str.data,
+                                        node->sn.str.len);
         if (rc != NJS_OK) {
             goto fail;
         }
@@ -726,7 +727,8 @@ njs_js_ext_shared_dict_incr(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
     double               value;
     ngx_int_t            rc;
     njs_str_t            key;
-    njs_value_t         *delta, *init;
+    ngx_msec_t           timeout;
+    njs_value_t         *delta, *init, *timeo;
     ngx_js_dict_t       *dict;
     ngx_shm_zone_t      *shm_zone;
     njs_opaque_value_t   lvalue;
@@ -765,7 +767,30 @@ njs_js_ext_shared_dict_incr(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         njs_value_number_set(init, 0);
     }
 
-    rc = ngx_js_dict_incr(shm_zone->data, &key, delta, init, &value);
+    timeo = njs_arg(args, nargs, 4);
+    if (!njs_value_is_undefined(timeo)) {
+        if (!njs_value_is_number(timeo)) {
+            njs_vm_type_error(vm, "timeout is not a number");
+            return NJS_ERROR;
+        }
+
+        if (!dict->timeout) {
+            njs_vm_type_error(vm, "shared dict must be declared with timeout");
+            return NJS_ERROR;
+        }
+
+        timeout = (ngx_msec_t) njs_value_number(timeo);
+
+        if (timeout < 1) {
+            njs_vm_type_error(vm, "timeout must be greater than or equal to 1");
+            return NJS_ERROR;
+        }
+
+    } else {
+        timeout = dict->timeout;
+    }
+
+    rc = ngx_js_dict_incr(shm_zone->data, &key, delta, init, &value, timeout);
     if (rc == NGX_ERROR) {
         njs_vm_error(vm, "failed to increment value in shared dict");
         return NJS_ERROR;
@@ -853,8 +878,8 @@ njs_js_ext_shared_dict_items(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
             goto fail;
         }
 
-        rc = njs_vm_value_string_set(vm, value, node->sn.str.data,
-                                     node->sn.str.len);
+        rc = njs_vm_value_string_create(vm, value, node->sn.str.data,
+                                        node->sn.str.len);
         if (rc != NJS_OK) {
             goto fail;
         }
@@ -896,8 +921,8 @@ njs_js_ext_shared_dict_name(njs_vm_t *vm, njs_object_prop_t *prop,
         return NJS_DECLINED;
     }
 
-    return njs_vm_value_string_set(vm, retval, shm_zone->shm.name.data,
-                                   shm_zone->shm.name.len);
+    return njs_vm_value_string_create(vm, retval, shm_zone->shm.name.data,
+                                      shm_zone->shm.name.len);
 }
 
 
@@ -936,7 +961,8 @@ njs_js_ext_shared_dict_set(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 {
     njs_str_t        key;
     ngx_int_t        rc;
-    njs_value_t     *value;
+    ngx_msec_t       timeout;
+    njs_value_t     *value, *timeo;
     ngx_js_dict_t   *dict;
     ngx_shm_zone_t  *shm_zone;
 
@@ -967,7 +993,30 @@ njs_js_ext_shared_dict_set(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         }
     }
 
-    rc = ngx_js_dict_set(vm, shm_zone->data, &key, value, flags);
+    timeo = njs_arg(args, nargs, 3);
+    if (!njs_value_is_undefined(timeo)) {
+        if (!njs_value_is_number(timeo)) {
+            njs_vm_type_error(vm, "timeout is not a number");
+            return NJS_ERROR;
+        }
+
+        if (!dict->timeout) {
+            njs_vm_type_error(vm, "shared dict must be declared with timeout");
+            return NJS_ERROR;
+        }
+
+        timeout = (ngx_msec_t) njs_value_number(timeo);
+
+        if (timeout < 1) {
+            njs_vm_type_error(vm, "timeout must be greater than or equal to 1");
+            return NJS_ERROR;
+        }
+
+    } else {
+        timeout = dict->timeout;
+    }
+
+    rc = ngx_js_dict_set(vm, shm_zone->data, &key, value, timeout, flags);
     if (rc == NGX_ERROR) {
         return NJS_ERROR;
     }
@@ -1063,7 +1112,7 @@ njs_js_ext_shared_dict_type(njs_vm_t *vm, njs_object_prop_t *prop,
         break;
     }
 
-    return njs_vm_value_string_set(vm, retval, type.start, type.length);
+    return njs_vm_value_string_create(vm, retval, type.start, type.length);
 }
 
 
@@ -1118,7 +1167,7 @@ ngx_js_dict_node_free(ngx_js_dict_t *dict, ngx_js_dict_node_t *node)
 
 static ngx_int_t
 ngx_js_dict_set(njs_vm_t *vm, ngx_js_dict_t *dict, njs_str_t *key,
-    njs_value_t *value, unsigned flags)
+    njs_value_t *value, ngx_msec_t timeout, unsigned flags)
 {
     ngx_msec_t           now;
     ngx_time_t          *tp;
@@ -1137,7 +1186,7 @@ ngx_js_dict_set(njs_vm_t *vm, ngx_js_dict_t *dict, njs_str_t *key,
             return NGX_DECLINED;
         }
 
-        if (ngx_js_dict_add(dict, key, value, now) != NGX_OK) {
+        if (ngx_js_dict_add(dict, key, value, timeout, now) != NGX_OK) {
             goto memory_error;
         }
 
@@ -1149,7 +1198,7 @@ ngx_js_dict_set(njs_vm_t *vm, ngx_js_dict_t *dict, njs_str_t *key,
             }
         }
 
-        if (ngx_js_dict_update(dict, node, value, now) != NGX_OK) {
+        if (ngx_js_dict_update(dict, node, value, timeout, now) != NGX_OK) {
             goto memory_error;
         }
     }
@@ -1170,7 +1219,7 @@ memory_error:
 
 static ngx_int_t
 ngx_js_dict_add(ngx_js_dict_t *dict, njs_str_t *key, njs_value_t *value,
-    ngx_msec_t now)
+    ngx_msec_t timeout, ngx_msec_t now)
 {
     size_t               n;
     uint32_t             hash;
@@ -1214,7 +1263,7 @@ ngx_js_dict_add(ngx_js_dict_t *dict, njs_str_t *key, njs_value_t *value,
     ngx_rbtree_insert(&dict->sh->rbtree, &node->sn.node);
 
     if (dict->timeout) {
-        node->expire.key = now + dict->timeout;
+        node->expire.key = now + timeout;
         ngx_rbtree_insert(&dict->sh->rbtree_expire, &node->expire);
     }
 
@@ -1224,7 +1273,7 @@ ngx_js_dict_add(ngx_js_dict_t *dict, njs_str_t *key, njs_value_t *value,
 
 static ngx_int_t
 ngx_js_dict_update(ngx_js_dict_t *dict, ngx_js_dict_node_t *node,
-    njs_value_t *value, ngx_msec_t now)
+    njs_value_t *value, ngx_msec_t timeout, ngx_msec_t now)
 {
     u_char     *p;
     njs_str_t   string;
@@ -1249,7 +1298,7 @@ ngx_js_dict_update(ngx_js_dict_t *dict, ngx_js_dict_node_t *node,
 
     if (dict->timeout) {
         ngx_rbtree_delete(&dict->sh->rbtree_expire, &node->expire);
-        node->expire.key = now + dict->timeout;
+        node->expire.key = now + timeout;
         ngx_rbtree_insert(&dict->sh->rbtree_expire, &node->expire);
     }
 
@@ -1306,7 +1355,7 @@ ngx_js_dict_delete(njs_vm_t *vm, ngx_js_dict_t *dict, njs_str_t *key,
 
 static ngx_int_t
 ngx_js_dict_incr(ngx_js_dict_t *dict, njs_str_t *key, njs_value_t *delta,
-    njs_value_t *init, double *value)
+    njs_value_t *init, double *value, ngx_msec_t timeout)
 {
     ngx_msec_t           now;
     ngx_time_t          *tp;
@@ -1322,7 +1371,7 @@ ngx_js_dict_incr(ngx_js_dict_t *dict, njs_str_t *key, njs_value_t *delta,
     if (node == NULL) {
         njs_value_number_set(init, njs_value_number(init)
                                    + njs_value_number(delta));
-        if (ngx_js_dict_add(dict, key, init, now) != NGX_OK) {
+        if (ngx_js_dict_add(dict, key, init, timeout, now) != NGX_OK) {
             ngx_rwlock_unlock(&dict->sh->rwlock);
             return NGX_ERROR;
         }
@@ -1335,7 +1384,7 @@ ngx_js_dict_incr(ngx_js_dict_t *dict, njs_str_t *key, njs_value_t *delta,
 
         if (dict->timeout) {
             ngx_rbtree_delete(&dict->sh->rbtree_expire, &node->expire);
-            node->expire.key = now + dict->timeout;
+            node->expire.key = now + timeout;
             ngx_rbtree_insert(&dict->sh->rbtree_expire, &node->expire);
         }
     }
@@ -1391,22 +1440,13 @@ ngx_js_dict_copy_value_locked(njs_vm_t *vm, ngx_js_dict_t *dict,
     ngx_js_dict_node_t *node, njs_value_t *retval)
 {
     njs_int_t   ret;
-    njs_str_t   string;
     ngx_uint_t  type;
-    ngx_pool_t  *pool;
 
     type = dict->type;
 
     if (type == NGX_JS_DICT_TYPE_STRING) {
-        pool = ngx_external_pool(vm, njs_vm_external_ptr(vm));
-
-        string.length = node->u.value.len;
-        string.start = ngx_pstrdup(pool, &node->u.value);
-        if (string.start == NULL) {
-            return NGX_ERROR;
-        }
-
-        ret = njs_vm_value_string_set(vm, retval, string.start, string.length);
+        ret = njs_vm_value_string_create(vm, retval, node->u.value.data,
+                                         node->u.value.len);
         if (ret != NJS_OK) {
             return NGX_ERROR;
         }
@@ -1493,8 +1533,8 @@ static njs_int_t
 ngx_js_dict_shared_error_name(njs_vm_t *vm, njs_object_prop_t *prop,
     njs_value_t *value, njs_value_t *setval, njs_value_t *retval)
 {
-    return njs_vm_value_string_set(vm, retval, (u_char *) "SharedMemoryError",
-                                   17);
+    return njs_vm_value_string_create(vm, retval,
+                                      (u_char *) "SharedMemoryError", 17);
 }
 
 
