@@ -25,12 +25,7 @@ typedef struct {
 static njs_int_t njs_global_this_prop_handler(njs_vm_t *vm,
     njs_object_prop_t *self, njs_value_t *global, njs_value_t *setval,
     njs_value_t *retval);
-static njs_arr_t *njs_vm_expression_completions(njs_vm_t *vm,
-    njs_str_t *expression);
-static njs_arr_t *njs_vm_global_var_completions(njs_vm_t *vm,
-	njs_str_t *expression);
-static njs_arr_t *njs_object_completions(njs_vm_t *vm, njs_value_t *object,
-    njs_str_t *expression);
+
 static njs_int_t njs_env_hash_init(njs_vm_t *vm, njs_lvlhsh_t *hash,
     char **environment);
 
@@ -258,8 +253,6 @@ njs_builtin_objects_create(njs_vm_t *vm)
     vm->global_object = shared->objects[0];
     vm->global_object.shared = 0;
 
-    njs_set_object(&vm->global_value, &vm->global_object);
-
     string_object = &shared->string_object;
     njs_lvlhsh_init(&string_object->hash);
     string_object->shared_hash = shared->string_instance_hash;
@@ -395,7 +388,7 @@ njs_builtin_traverse(njs_vm_t *vm, njs_traverse_t *traverse, void *data)
         return NJS_ERROR;
     }
 
-    ret = njs_string_new(vm, &prop->name, buf, p - buf, 0);
+    ret = njs_string_create(vm, &prop->name, buf, p - buf);
     if (njs_slow_path(ret != NJS_OK)) {
         return ret;
     }
@@ -417,293 +410,6 @@ njs_builtin_traverse(njs_vm_t *vm, njs_traverse_t *traverse, void *data)
 }
 
 
-static njs_arr_t *
-njs_builtin_completions(njs_vm_t *vm)
-{
-    njs_arr_t                *array;
-    njs_str_t                *completion;
-    njs_int_t                ret;
-    njs_lvlhsh_each_t        lhe;
-    njs_builtin_traverse_t   ctx;
-    const njs_object_prop_t  *prop;
-
-    array = njs_arr_create(vm->mem_pool, 64, sizeof(njs_str_t));
-    if (njs_slow_path(array == NULL)) {
-        return NULL;
-    }
-
-    ret = njs_lexer_keywords(array);
-    if (njs_slow_path(ret != NJS_OK)) {
-        return NULL;
-    }
-
-    /* Global object completions. */
-
-    ctx.type = NJS_BUILTIN_TRAVERSE_KEYS;
-    njs_lvlhsh_init(&ctx.keys);
-
-    ret = njs_object_traverse(vm, &vm->global_object, &ctx,
-                              njs_builtin_traverse);
-    if (njs_slow_path(ret != NJS_OK)) {
-        return NULL;
-    }
-
-    njs_lvlhsh_each_init(&lhe, &njs_object_hash_proto);
-
-    for ( ;; ) {
-        prop = njs_lvlhsh_each(&ctx.keys, &lhe);
-
-        if (prop == NULL) {
-            break;
-        }
-
-        completion = njs_arr_add(array);
-        if (njs_slow_path(completion == NULL)) {
-            return NULL;
-        }
-
-        njs_string_get(&prop->name, completion);
-    }
-
-    return array;
-}
-
-
-void *
-njs_vm_completions(njs_vm_t *vm, njs_str_t *expression)
-{
-    u_char  *p, *end;
-
-    if (expression == NULL) {
-        return njs_builtin_completions(vm);
-    }
-
-    p = expression->start;
-    end = p + expression->length;
-
-    while (p < end && *p != '.') { p++; }
-
-    if (p == end) {
-        return njs_vm_global_var_completions(vm, expression);
-    }
-
-    return njs_vm_expression_completions(vm, expression);
-}
-
-
-static njs_arr_t *
-njs_vm_global_var_completions(njs_vm_t *vm, njs_str_t *expression)
-{
-    njs_str_t                *completion;
-    njs_arr_t                *array;
-    njs_rbtree_t             *variables;
-    njs_rbtree_node_t        *node;
-    njs_variable_node_t      *vnode;
-    const njs_lexer_entry_t  *lex_entry;
-
-    variables = (vm->global_scope != NULL) ? &vm->global_scope->variables
-                                           : NULL;
-    if (njs_slow_path(variables == NULL)) {
-        return NULL;
-    }
-
-    array = njs_arr_create(vm->mem_pool, 8, sizeof(njs_str_t));
-    if (njs_slow_path(array == NULL)) {
-        return NULL;
-    }
-
-    node = njs_rbtree_min(variables);
-
-    while (njs_rbtree_is_there_successor(variables, node)) {
-        vnode = (njs_variable_node_t *) node;
-
-        node = njs_rbtree_node_successor(variables, node);
-
-        lex_entry = njs_lexer_entry(vnode->key);
-        if (lex_entry == NULL) {
-            continue;
-        }
-
-        if (lex_entry->name.length >= expression->length
-            && njs_strncmp(expression->start, lex_entry->name.start,
-                           expression->length) == 0)
-        {
-            completion = njs_arr_add(array);
-            if (njs_slow_path(completion == NULL)) {
-                return NULL;
-            }
-
-            *completion = lex_entry->name;
-        }
-    }
-
-    return array;
-}
-
-
-static njs_arr_t *
-njs_vm_expression_completions(njs_vm_t *vm, njs_str_t *expression)
-{
-    u_char               *p, *end;
-    njs_int_t            ret;
-    njs_value_t          *value;
-    njs_variable_t       *var;
-    njs_rbtree_node_t    *node;
-    njs_object_prop_t    *prop;
-    njs_lvlhsh_query_t   lhq;
-    njs_variable_node_t  var_node;
-
-    p = expression->start;
-    end = p + expression->length;
-
-    lhq.key.start = p;
-
-    while (p < end && *p != '.') { p++; }
-
-    lhq.proto = &njs_lexer_hash_proto;
-    lhq.key.length = p - lhq.key.start;
-    lhq.key_hash = njs_djb_hash(lhq.key.start, lhq.key.length);
-
-    ret = njs_lvlhsh_find(&vm->shared->keywords_hash, &lhq);
-    if (njs_slow_path(ret != NJS_OK)) {
-        return NULL;
-    }
-
-    var_node.key = (uintptr_t) lhq.value;
-
-    node = njs_rbtree_find(&vm->global_scope->variables, &var_node.node);
-    if (njs_slow_path(node == NULL)) {
-        return NULL;
-    }
-
-    var = ((njs_variable_node_t *) node)->variable;
-    value = njs_scope_value(vm, var->index);
-
-    if (!njs_is_object(value)) {
-        return NULL;
-    }
-
-    lhq.proto = &njs_object_hash_proto;
-
-    for ( ;; ) {
-
-        if (p == end) {
-            break;
-        }
-
-        lhq.key.start = ++p;
-
-        while (p < end && *p != '.') { p++; }
-
-        lhq.key.length = p - lhq.key.start;
-        lhq.key_hash = njs_djb_hash(lhq.key.start, lhq.key.length);
-
-        ret = njs_lvlhsh_find(njs_object_hash(value), &lhq);
-        if (njs_slow_path(ret != NJS_OK)) {
-            if (ret == NJS_DECLINED) {
-                break;
-            }
-
-            return NULL;
-        }
-
-        prop = lhq.value;
-
-        if (njs_is_accessor_descriptor(prop) ||
-            !njs_is_object(njs_prop_value(prop)))
-        {
-            return NULL;
-        }
-
-        value = njs_prop_value(prop);
-    }
-
-    return njs_object_completions(vm, value, expression);
-}
-
-
-static njs_arr_t *
-njs_object_completions(njs_vm_t *vm, njs_value_t *object, njs_str_t *expression)
-{
-    u_char            *prefix;
-    double            num;
-    size_t            len;
-    njs_arr_t         *array;
-    njs_str_t         *completion, key;
-    njs_uint_t        n;
-    njs_array_t       *keys;
-    njs_value_type_t  type;
-
-    prefix = expression->start + expression->length;
-
-    while (prefix > expression->start && *prefix != '.') {
-        prefix--;
-    }
-
-    prefix++;
-    len = expression->length - (prefix - expression->start);
-
-    array = NULL;
-    type = object->type;
-
-    if (type == NJS_ARRAY || type == NJS_TYPED_ARRAY) {
-        object->type = NJS_OBJECT;
-    }
-
-    keys = njs_value_enumerate(vm, object, NJS_ENUM_KEYS, NJS_ENUM_STRING, 1);
-    if (njs_slow_path(keys == NULL)) {
-        goto done;
-    }
-
-    array = njs_arr_create(vm->mem_pool, 8, sizeof(njs_str_t));
-    if (njs_slow_path(array == NULL)) {
-        goto done;
-    }
-
-    for (n = 0; n < keys->length; n++) {
-        njs_string_get(&keys->start[n], &key);
-
-        if (len != 0
-            && njs_strncmp(key.start, prefix, njs_min(len, key.length)) != 0)
-        {
-            continue;
-        }
-
-        num = njs_key_to_index(&keys->start[n]);
-
-        if (!njs_key_is_integer_index(num, &keys->start[n])) {
-            completion = njs_arr_add(array);
-            if (njs_slow_path(completion == NULL)) {
-                njs_arr_destroy(array);
-                array = NULL;
-                goto done;
-            }
-
-            completion->length = (prefix - expression->start) + key.length + 1;
-            completion->start = njs_mp_alloc(vm->mem_pool, completion->length);
-            if (njs_slow_path(completion->start == NULL)) {
-                njs_arr_destroy(array);
-                array = NULL;
-                goto done;
-            }
-
-            njs_sprintf(completion->start,
-                        completion->start + completion->length,
-                        "%*s%V%Z", prefix - expression->start,
-                        expression->start, &key);
-        }
-    }
-
-done:
-
-    if (type == NJS_ARRAY || type == NJS_TYPED_ARRAY) {
-        object->type = type;
-    }
-
-    return array;
-}
-
-
 typedef struct {
     njs_str_t               name;
     njs_function_native_t   native;
@@ -753,7 +459,7 @@ njs_builtin_match_native_function(njs_vm_t *vm, njs_function_t *function,
 
     ctx.match = njs_str_value("");
 
-    ret = njs_object_traverse(vm, &vm->global_object, &ctx,
+    ret = njs_object_traverse(vm, njs_object(&vm->global_value), &ctx,
                               njs_builtin_traverse);
 
     if (ret == NJS_DONE) {
@@ -884,7 +590,7 @@ njs_ext_dump(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         return NJS_ERROR;
     }
 
-    return njs_string_new(vm, retval, str.start, str.length, 0);
+    return njs_string_create(vm, retval, str.start, str.length);
 }
 
 
@@ -1259,12 +965,6 @@ static const njs_object_prop_t  njs_global_this_object_properties[] =
 
     NJS_DECLARE_PROP_NATIVE("eval", njs_eval_function, 1, 0),
 
-    NJS_DECLARE_PROP_NATIVE("setTimeout", njs_set_timeout, 2, 0),
-
-    NJS_DECLARE_PROP_NATIVE("setImmediate", njs_set_immediate, 4, 0),
-
-    NJS_DECLARE_PROP_NATIVE("clearTimeout", njs_clear_timeout, 1, 0),
-
     NJS_DECLARE_PROP_NATIVE("require", njs_module_require, 1, 0),
 
     /* Global objects. */
@@ -1454,6 +1154,9 @@ static const njs_object_prop_t  njs_njs_object_properties[] =
         .configurable = 1,
     },
 
+    NJS_DECLARE_PROP_VALUE("engine", njs_string("njs"),
+                           NJS_OBJECT_PROP_VALUE_EC),
+
     NJS_DECLARE_PROP_VALUE("version", njs_string(NJS_VERSION),
                            NJS_OBJECT_PROP_VALUE_EC),
 
@@ -1498,8 +1201,8 @@ njs_process_object_argv(njs_vm_t *vm, njs_object_prop_t *pr,
     i = 0;
 
     for (arg = vm->options.argv; i < vm->options.argc; arg++) {
-        njs_string_set(vm, &argv->start[i++], (u_char *) *arg,
-                       njs_strlen(*arg));
+        njs_string_create(vm, &argv->start[i++], (u_char *) *arg,
+                          njs_strlen(*arg));
     }
 
     prop = njs_object_prop_alloc(vm, &argv_string, &njs_value_undefined, 1);
@@ -1538,7 +1241,7 @@ njs_env_hash_init(njs_vm_t *vm, njs_lvlhsh_t *hash, char **environment)
     uint32_t            cp;
     njs_int_t           ret;
     const u_char        *val, *entry, *s, *end;
-    njs_object_prop_t   *prop, *prev;
+    njs_object_prop_t   *prop;
     njs_string_prop_t   string;
     njs_lvlhsh_query_t  lhq;
 
@@ -1562,7 +1265,7 @@ njs_env_hash_init(njs_vm_t *vm, njs_lvlhsh_t *hash, char **environment)
             continue;
         }
 
-        ret = njs_string_create(vm, &prop->name, (char *) entry, val - entry);
+        ret = njs_string_create(vm, &prop->name, entry, val - entry);
         if (njs_slow_path(ret != NJS_OK)) {
             return NJS_ERROR;
         }
@@ -1582,8 +1285,7 @@ njs_env_hash_init(njs_vm_t *vm, njs_lvlhsh_t *hash, char **environment)
 
         val++;
 
-        ret = njs_string_create(vm, njs_prop_value(prop), (char *) val,
-                                njs_strlen(val));
+        ret = njs_string_create(vm, njs_prop_value(prop), val, njs_strlen(val));
         if (njs_slow_path(ret != NJS_OK)) {
             return NJS_ERROR;
         }
@@ -1605,13 +1307,6 @@ njs_env_hash_init(njs_vm_t *vm, njs_lvlhsh_t *hash, char **environment)
              * Always using the first element among the duplicates
              * and ignoring the rest.
              */
-
-            prev = lhq.value;
-
-            if (!njs_values_same(njs_prop_value(prop), njs_prop_value(prev))) {
-                njs_vm_warn(vm, "environment variable \"%V\" has more than one"
-                            " value\n", &lhq.key);
-            }
         }
     }
 
