@@ -310,7 +310,6 @@ njs_function_rest_parameters_init(njs_vm_t *vm, njs_native_frame_t *frame)
         return NJS_ERROR;
     }
 
-    /* GC: retain. */
     njs_set_array(rest_arguments, array);
 
     vm->top_frame->local[n] = rest_arguments;
@@ -436,7 +435,7 @@ njs_function_lambda_frame(njs_vm_t *vm, njs_function_t *function,
     if (njs_slow_path(function->global_this
                       && njs_is_null_or_undefined(this)))
     {
-        njs_set_object(native_frame->local[0], &vm->global_object);
+        njs_value_assign(native_frame->local[0], &vm->global_value);
     }
 
     /* Copy arguments. */
@@ -1033,7 +1032,7 @@ njs_function_constructor(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
         goto fail;
     }
 
-    njs_chb_init(&chain, vm->mem_pool);
+    NJS_CHB_MP_INIT(&chain, vm);
 
     if (async) {
         njs_chb_append_literal(&chain, "(async function(");
@@ -1419,8 +1418,6 @@ njs_function_prototype_bind(njs_vm_t *vm, njs_value_t *args, njs_uint_t nargs,
 
     function->bound = values;
 
-    /* GC: ? retain args. */
-
     memcpy(values, args, size);
 
     njs_set_function(retval, function);
