@@ -1388,7 +1388,7 @@ njs_typed_array_prototype_index_of(njs_vm_t *vm, njs_value_t *args,
 
     v = njs_number(njs_argument(args, 1));
 
-    i64 = v;
+    i64 = njs_unsafe_cast_double_to_int64(v);
     integer = (v == i64);
 
     buffer = array->buffer;
@@ -2035,6 +2035,11 @@ njs_typed_array_prototype_sort(njs_vm_t *vm, njs_value_t *args,
     }
 
     njs_qsort(base, length, element_size, cmp, &ctx);
+
+    if (njs_slow_path(ctx.exception)) {
+        return NJS_ERROR;
+    }
+
     if (ctx.function != NULL) {
         if (&buffer->u.u8[array->offset * element_size] == orig) {
             memcpy(orig, base, length * element_size);
@@ -2043,10 +2048,6 @@ njs_typed_array_prototype_sort(njs_vm_t *vm, njs_value_t *args,
         njs_mp_free(vm->mem_pool, base);
     }
 
-    if (njs_slow_path(ctx.exception)) {
-        return NJS_ERROR;
-    }
-
     njs_set_typed_array(retval, array);
 
     return NJS_OK;
@@ -2142,7 +2143,7 @@ njs_typed_array_prototype_join(njs_vm_t *vm, njs_value_t *args,
         return NJS_ERROR;
     }
 
-    njs_chb_init(&chain, vm->mem_pool);
+    NJS_CHB_MP_INIT(&chain, vm);
 
     length = njs_typed_array_to_chain(vm, &chain, array, separator);
     size = njs_chb_size(&chain);
