@@ -8,7 +8,6 @@
 #include <njs_main.h>
 
 
-static njs_int_t njs_vm_handle_events(njs_vm_t *vm);
 static njs_int_t njs_vm_protos_init(njs_vm_t *vm, njs_value_t *global);
 
 
@@ -25,7 +24,6 @@ njs_vm_opt_init(njs_vm_opt_t *options)
 {
     njs_memzero(options, sizeof(njs_vm_opt_t));
 
-    options->log_level = NJS_LOG_LEVEL_INFO;
     options->max_stack_size = NJS_MAX_STACK_SIZE;
 }
 
@@ -198,27 +196,10 @@ njs_vm_ctor_push(njs_vm_t *vm)
 void
 njs_vm_destroy(njs_vm_t *vm)
 {
-    njs_event_t        *event;
-    njs_lvlhsh_each_t  lhe;
-
     if (vm->hooks[NJS_HOOK_EXIT] != NULL) {
         (void) njs_vm_call(vm, vm->hooks[NJS_HOOK_EXIT], NULL, 0);
     }
 
-    if (njs_waiting_events(vm)) {
-        njs_lvlhsh_each_init(&lhe, &njs_event_hash_proto);
-
-        for ( ;; ) {
-            event = njs_lvlhsh_each(&vm->events_hash, &lhe);
-
-            if (event == NULL) {
-                break;
-            }
-
-            njs_del_event(vm, event, NJS_EVENT_RELEASE);
-        }
-    }
-
     njs_mp_destroy(vm->mem_pool);
 }
 
@@ -252,7 +233,7 @@ njs_vm_compile(njs_vm_t *vm, u_char **start, u_char *end)
     }
 
     if (njs_slow_path(vm->options.ast)) {
-        njs_chb_init(&chain, vm->mem_pool);
+        NJS_CHB_MP_INIT(&chain, vm);
         ret = njs_parser_serialize_ast(parser.node, &chain);
         if (njs_slow_path(ret == NJS_ERROR)) {
             return ret;
@@ -444,8 +425,6 @@ njs_vm_clone(njs_vm_t *vm, njs_external_ptr_t external)
 
     nvm->levels[NJS_LEVEL_GLOBAL] = global;
 
-    njs_set_object(&nvm->global_value, &nvm->global_object);
-
     /* globalThis and this */
     njs_scope_value_set(nvm, njs_scope_global_this_index(), &nvm->global_value);
 
@@ -487,12 +466,10 @@ njs_vm_runtime_init(njs_vm_t *vm)
     njs_lvlhsh_init(&vm->values_hash);
     njs_lvlhsh_init(&vm->keywords_hash);
     njs_lvlhsh_init(&vm->modules_hash);
-    njs_lvlhsh_init(&vm->events_hash);
 
     njs_rbtree_init(&vm->global_symbols, njs_symbol_rbtree_cmp);
 
-    njs_queue_init(&vm->posted_events);
-    njs_queue_init(&vm->promise_events);
+    njs_queue_init(&vm->jobs);
 
     return NJS_OK;
 }
@@ -640,102 +617,45 @@ njs_vm_scopes_restore(njs_vm_t *vm, njs_native_frame_t *native)
 }
 
 
-njs_vm_event_t
-njs_vm_add_event(njs_vm_t *vm, njs_function_t *function, njs_uint_t once,
-    njs_host_event_t host_ev, njs_event_destructor_t destructor)
+njs_int_t
+njs_vm_pending(njs_vm_t *vm)
 {
-    njs_event_t  *event;
-
-    event = njs_mp_alloc(vm->mem_pool, sizeof(njs_event_t));
-    if (njs_slow_path(event == NULL)) {
-        return NULL;
-    }
-
-    event->host_event = host_ev;
-    event->destructor = destructor;
-    event->function = function;
-    event->once = once;
-    event->posted = 0;
-    event->nargs = 0;
-    event->args = NULL;
-
-    if (njs_add_event(vm, event) != NJS_OK) {
-        return NULL;
-    }
-
-    return event;
-}
-
-
-void
-njs_vm_del_event(njs_vm_t *vm, njs_vm_event_t vm_event)
-{
-    njs_event_t  *event;
-
-    event = (njs_event_t *) vm_event;
-
-    njs_del_event(vm, event, NJS_EVENT_RELEASE | NJS_EVENT_DELETE);
+    return !njs_queue_is_empty(&(vm)->jobs);
 }
 
 
 njs_int_t
-njs_vm_waiting(njs_vm_t *vm)
-{
-    return njs_waiting_events(vm);
-}
-
-
-njs_int_t
-njs_vm_posted(njs_vm_t *vm)
-{
-    return njs_posted_events(vm) || njs_promise_events(vm);
-}
-
-
-njs_int_t
-njs_vm_unhandled_rejection(njs_vm_t *vm)
-{
-    return vm->options.unhandled_rejection
-             == NJS_VM_OPT_UNHANDLED_REJECTION_THROW
-           && vm->promise_reason != NULL
-           && vm->promise_reason->length != 0;
-}
-
-
-njs_int_t
-njs_vm_post_event(njs_vm_t *vm, njs_vm_event_t vm_event,
+njs_vm_enqueue_job(njs_vm_t *vm, njs_function_t *function,
     const njs_value_t *args, njs_uint_t nargs)
 {
     njs_event_t  *event;
 
-    event = (njs_event_t *) vm_event;
+    event = njs_mp_zalloc(vm->mem_pool, sizeof(njs_event_t));
+    if (njs_slow_path(event == NULL)) {
+        njs_memory_error(vm);
+        return NJS_ERROR;
+    }
 
-    if (nargs != 0 && !event->posted) {
-        event->nargs = nargs;
+    event->function = function;
+
+    if (nargs != 0) {
         event->args = njs_mp_alloc(vm->mem_pool, sizeof(njs_value_t) * nargs);
         if (njs_slow_path(event->args == NULL)) {
+            njs_memory_error(vm);
             return NJS_ERROR;
         }
 
         memcpy(event->args, args, sizeof(njs_value_t) * nargs);
+
+        event->nargs = nargs;
     }
 
-    if (!event->posted) {
-        event->posted = 1;
-        njs_queue_insert_tail(&vm->posted_events, &event->link);
-    }
+    njs_queue_insert_tail(&vm->jobs, &event->link);
 
     return NJS_OK;
 }
 
 
-njs_int_t
-njs_vm_run(njs_vm_t *vm)
-{
-    return njs_vm_handle_events(vm);
-}
-
-
 njs_int_t
 njs_vm_start(njs_vm_t *vm, njs_value_t *retval)
 {
@@ -747,103 +667,50 @@ njs_vm_start(njs_vm_t *vm, njs_value_t *retval)
 }
 
 
-static njs_int_t
-njs_vm_handle_events(njs_vm_t *vm)
+njs_int_t
+njs_vm_execute_pending_job(njs_vm_t *vm)
 {
     njs_int_t         ret;
-    njs_str_t         str;
-    njs_value_t       string;
     njs_event_t       *ev;
-    njs_queue_t       *promise_events, *posted_events;
+    njs_queue_t       *jobs;
     njs_queue_link_t  *link;
 
-    promise_events = &vm->promise_events;
-    posted_events = &vm->posted_events;
+    jobs = &vm->jobs;
 
-    do {
-        for ( ;; ) {
-            link = njs_queue_first(promise_events);
+    link = njs_queue_first(jobs);
 
-            if (link == njs_queue_tail(promise_events)) {
-                break;
-            }
+    if (link == njs_queue_tail(jobs)) {
+        return NJS_OK;
+    }
 
-            ev = njs_queue_link_data(link, njs_event_t, link);
+    ev = njs_queue_link_data(link, njs_event_t, link);
 
-            njs_queue_remove(&ev->link);
+    njs_queue_remove(&ev->link);
 
-            ret = njs_vm_call(vm, ev->function, ev->args, ev->nargs);
-            if (njs_slow_path(ret == NJS_ERROR)) {
-                return ret;
-            }
-        }
+    ret = njs_vm_call(vm, ev->function, ev->args, ev->nargs);
+    if (ret == NJS_ERROR) {
+        return ret;
+    }
 
-        if (njs_vm_unhandled_rejection(vm)) {
-            njs_value_assign(&string, &vm->promise_reason->start[0]);
-            ret = njs_value_to_string(vm, &string, &string);
-            if (njs_slow_path(ret != NJS_OK)) {
-                return ret;
-            }
-
-            njs_string_get(&string, &str);
-            njs_vm_error(vm, "unhandled promise rejection: %V", &str);
-
-            njs_mp_free(vm->mem_pool, vm->promise_reason);
-            vm->promise_reason = NULL;
-
-            return NJS_ERROR;
-        }
-
-        for ( ;; ) {
-            link = njs_queue_first(posted_events);
-
-            if (link == njs_queue_tail(posted_events)) {
-                break;
-            }
-
-            ev = njs_queue_link_data(link, njs_event_t, link);
-
-            if (ev->once) {
-                njs_del_event(vm, ev, NJS_EVENT_RELEASE | NJS_EVENT_DELETE);
-
-            } else {
-                ev->posted = 0;
-                njs_queue_remove(&ev->link);
-            }
-
-            ret = njs_vm_call(vm, ev->function, ev->args, ev->nargs);
-
-            if (ret == NJS_ERROR) {
-                return ret;
-            }
-        }
-
-    } while (!njs_queue_is_empty(promise_events));
-
-    return njs_vm_pending(vm) ? NJS_AGAIN : NJS_OK;
+    return 1;
 }
 
 
-njs_int_t
-njs_vm_add_path(njs_vm_t *vm, const njs_str_t *path)
+void
+njs_vm_set_module_loader(njs_vm_t *vm, njs_module_loader_t module_loader,
+    void *opaque)
 {
-    njs_str_t  *item;
+    vm->module_loader = module_loader;
+    vm->module_loader_opaque = opaque;
+}
 
-    if (vm->paths == NULL) {
-        vm->paths = njs_arr_create(vm->mem_pool, 4, sizeof(njs_str_t));
-        if (njs_slow_path(vm->paths == NULL)) {
-            return NJS_ERROR;
-        }
-    }
 
-    item = njs_arr_add(vm->paths);
-    if (njs_slow_path(item == NULL)) {
-        return NJS_ERROR;
-    }
-
-    *item = *path;
-
-    return NJS_OK;
+void
+njs_vm_set_rejection_tracker(njs_vm_t *vm,
+    njs_rejection_tracker_t rejection_tracker, void *opaque)
+{
+    vm->rejection_tracker = rejection_tracker;
+    vm->rejection_tracker_opaque = opaque;
 }
 
 
@@ -946,6 +813,14 @@ njs_vm_error3(njs_vm_t *vm, unsigned type, const char *fmt, ...)
 }
 
 
+njs_int_t
+njs_vm_global(njs_vm_t *vm, njs_value_t *retval)
+{
+    njs_value_assign(retval, &vm->global_value);
+    return NJS_OK;
+}
+
+
 njs_int_t
 njs_vm_value(njs_vm_t *vm, const njs_str_t *path, njs_value_t *retval)
 {
@@ -957,7 +832,7 @@ njs_vm_value(njs_vm_t *vm, const njs_str_t *path, njs_value_t *retval)
     start = path->start;
     end = start + path->length;
 
-    njs_set_object(&value, &vm->global_object);
+    njs_value_assign(&value, &vm->global_value);
 
     for ( ;; ) {
         p = njs_strlchr(start, end, '.');
@@ -968,7 +843,7 @@ njs_vm_value(njs_vm_t *vm, const njs_str_t *path, njs_value_t *retval)
             return NJS_ERROR;
         }
 
-        ret = njs_string_set(vm, &key, start, size);
+        ret = njs_string_create(vm, &key, start, size);
         if (njs_slow_path(ret != NJS_OK)) {
             return NJS_ERROR;
         }
@@ -999,7 +874,7 @@ njs_vm_bind2(njs_vm_t *vm, const njs_str_t *var_name, njs_object_prop_t *prop,
     njs_lvlhsh_t        *hash;
     njs_lvlhsh_query_t  lhq;
 
-    ret = njs_string_new(vm, &prop->name, var_name->start, var_name->length, 0);
+    ret = njs_string_create(vm, &prop->name, var_name->start, var_name->length);
     if (njs_slow_path(ret != NJS_OK)) {
         return NJS_ERROR;
     }
@@ -1070,14 +945,6 @@ njs_value_string_get(njs_value_t *value, njs_str_t *dst)
 }
 
 
-njs_int_t
-njs_vm_value_string_set(njs_vm_t *vm, njs_value_t *value, const u_char *start,
-    uint32_t size)
-{
-    return njs_string_set(vm, value, start, size);
-}
-
-
 njs_int_t
 njs_vm_value_array_buffer_set(njs_vm_t *vm, njs_value_t *value,
     const u_char *start, uint32_t size)
@@ -1106,18 +973,11 @@ njs_vm_value_buffer_set(njs_vm_t *vm, njs_value_t *value, const u_char *start,
 }
 
 
-u_char *
-njs_vm_value_string_alloc(njs_vm_t *vm, njs_value_t *value, uint32_t size)
-{
-    return njs_string_alloc(vm, value, size, 0);
-}
-
-
 njs_int_t
 njs_vm_value_string_create(njs_vm_t *vm, njs_value_t *value,
     const u_char *start, uint32_t size)
 {
-    return njs_string_create(vm, value, (const char *) start, size);
+    return njs_string_create(vm, value, start, size);
 }
 
 
@@ -1178,30 +1038,6 @@ njs_vm_memory_error(njs_vm_t *vm)
 }
 
 
-njs_noinline void
-njs_vm_logger(njs_vm_t *vm, njs_log_level_t level, const char *fmt, ...)
-{
-    u_char        *p;
-    va_list       args;
-    njs_logger_t  logger;
-    u_char        buf[32768];
-
-    if (vm->options.ops == NULL) {
-        return;
-    }
-
-    logger = vm->options.ops->logger;
-
-    if (logger != NULL && vm->options.log_level >= level) {
-        va_start(args, fmt);
-        p = njs_vsprintf(buf, buf + sizeof(buf), fmt, args);
-        va_end(args);
-
-        logger(vm, vm->external, level, buf, p - buf);
-    }
-}
-
-
 njs_int_t
 njs_vm_value_string(njs_vm_t *vm, njs_str_t *dst, njs_value_t *src)
 {
@@ -1263,6 +1099,94 @@ njs_vm_exception_string(njs_vm_t *vm, njs_str_t *dst)
 }
 
 
+njs_value_t *
+njs_vm_value_enumerate(njs_vm_t *vm, njs_value_t *value, uint32_t flags,
+    njs_value_t *retval)
+{
+    ssize_t                  length;
+    njs_int_t                ret;
+    njs_value_t              *val;
+    njs_array_t              *keys;
+    njs_rbtree_t             *variables;
+    njs_rbtree_node_t        *rb_node;
+    njs_variable_node_t      *node;
+    const njs_lexer_entry_t  *lex_entry;
+
+    static const njs_str_t  njs_this_str = njs_str("this");
+
+    keys = njs_value_enumerate(vm, value, flags);
+    if (njs_slow_path(keys == NULL)) {
+        return NULL;
+    }
+
+    if (!njs_values_same(value, &vm->global_value)
+        || vm->global_scope == NULL)
+    {
+        goto done;
+    }
+
+    /* TODO: workaround for values in global object. */
+
+    variables = &vm->global_scope->variables;
+    rb_node = njs_rbtree_min(variables);
+
+    while (njs_rbtree_is_there_successor(variables, rb_node)) {
+        node = (njs_variable_node_t *) rb_node;
+
+        lex_entry = njs_lexer_entry(node->variable->unique_id);
+        if (njs_slow_path(lex_entry == NULL)) {
+            return NULL;
+        }
+
+        if (njs_strstr_eq(&lex_entry->name, &njs_this_str)) {
+            rb_node = njs_rbtree_node_successor(variables, rb_node);
+            continue;
+        }
+
+        length = njs_utf8_length(lex_entry->name.start, lex_entry->name.length);
+        if (njs_slow_path(length < 0)) {
+            return NULL;
+        }
+
+        val = njs_array_push(vm, keys);
+        if (njs_slow_path(value == NULL)) {
+            return NULL;
+        }
+
+        ret = njs_string_new(vm, val, lex_entry->name.start,
+                             lex_entry->name.length, length);
+        if (njs_slow_path(ret != NJS_OK)) {
+            return NULL;
+        }
+
+        rb_node = njs_rbtree_node_successor(variables, rb_node);
+    }
+
+done:
+
+     njs_set_array(retval, keys);
+
+     return retval;
+}
+
+
+njs_value_t *
+njs_vm_value_own_enumerate(njs_vm_t *vm, njs_value_t *value, uint32_t flags,
+    njs_value_t *retval)
+{
+    njs_array_t  *keys;
+
+    keys = njs_value_own_enumerate(vm, value, flags);
+    if (njs_slow_path(keys == NULL)) {
+        return NULL;
+    }
+
+    njs_set_array(retval, keys);
+
+    return retval;
+}
+
+
 njs_int_t
 njs_vm_object_alloc(njs_vm_t *vm, njs_value_t *retval, ...)
 {
@@ -1337,8 +1261,8 @@ njs_vm_object_keys(njs_vm_t *vm, njs_value_t *value, njs_value_t *retval)
 {
     njs_array_t  *keys;
 
-    keys = njs_value_own_enumerate(vm, value, NJS_ENUM_KEYS,
-                                   NJS_ENUM_STRING, 0);
+    keys = njs_value_own_enumerate(vm, value, NJS_ENUM_KEYS | NJS_ENUM_STRING
+                                   | NJS_ENUM_ENUMERABLE_ONLY);
     if (njs_slow_path(keys == NULL)) {
         return NULL;
     }
@@ -1349,6 +1273,19 @@ njs_vm_object_keys(njs_vm_t *vm, njs_value_t *value, njs_value_t *retval)
 }
 
 
+njs_int_t
+njs_vm_prototype(njs_vm_t *vm, njs_value_t *value, njs_value_t *retval)
+{
+    njs_value_t arguments[2];
+
+    njs_set_undefined(&arguments[0]);
+    njs_value_assign(&arguments[1], value);
+
+    return njs_object_get_prototype_of(vm, njs_value_arg(&arguments), 2, 0,
+                                       retval);
+}
+
+
 njs_int_t
 njs_vm_array_alloc(njs_vm_t *vm, njs_value_t *retval, uint32_t spare)
 {
@@ -1390,7 +1327,7 @@ njs_vm_object_prop(njs_vm_t *vm, njs_value_t *value, const njs_str_t *prop,
         return NULL;
     }
 
-    ret = njs_vm_value_string_set(vm, &key, prop->start, prop->length);
+    ret = njs_vm_value_string_create(vm, &key, prop->start, prop->length);
     if (njs_slow_path(ret != NJS_OK)) {
         return NULL;
     }
@@ -1416,7 +1353,7 @@ njs_vm_object_prop_set(njs_vm_t *vm, njs_value_t *value, const njs_str_t *prop,
         return NJS_ERROR;
     }
 
-    ret = njs_vm_value_string_set(vm, &key, prop->start, prop->length);
+    ret = njs_vm_value_string_create(vm, &key, prop->start, prop->length);
     if (njs_slow_path(ret != NJS_OK)) {
         return NJS_ERROR;
     }
@@ -1665,6 +1602,12 @@ njs_vm_value_to_bytes(njs_vm_t *vm, njs_str_t *dst, njs_value_t *src)
         }
 
         if (njs_slow_path(njs_is_detached_buffer(buffer))) {
+            if (length == 0) {
+                dst->length = 0;
+                dst->start = NULL;
+                return NJS_OK;
+            }
+
             njs_type_error(vm, "detached buffer");
             return NJS_ERROR;
         }
@@ -1710,47 +1653,6 @@ njs_vm_string_compare(const njs_value_t *v1, const njs_value_t *v2)
 }
 
 
-njs_int_t
-njs_vm_value_string_copy(njs_vm_t *vm, njs_str_t *retval,
-    njs_value_t *value, uintptr_t *next)
-{
-    uintptr_t    n;
-    njs_array_t  *array;
-
-    switch (value->type) {
-
-    case NJS_STRING:
-        if (*next != 0) {
-            return NJS_DECLINED;
-        }
-
-        *next = 1;
-        break;
-
-    case NJS_ARRAY:
-        array = njs_array(value);
-
-        do {
-            n = (*next)++;
-
-            if (n == array->length) {
-                return NJS_DECLINED;
-            }
-
-            value = &array->start[n];
-
-        } while (!njs_is_valid(value));
-
-        break;
-
-    default:
-        return NJS_ERROR;
-    }
-
-    return njs_vm_value_to_string(vm, retval, value);
-}
-
-
 void *
 njs_lvlhsh_alloc(void *data, size_t size)
 {
